//===========================================================================
/*
    CS277 - Experimental Haptics
    Spring 2014, Stanford University

    This class encapsulates the visual and haptic rendering for an implicit
    surface.  It inherits the regular CHAI3D cMesh class, and taps into
    an external implementation of the Marching Cubes algorithm to create
    a triangle mesh from the implicit surface function.

    Your job is to implement the method that tracks the position of the
    proxy as the tool moves and interacts with the object, as described by
    the implicit surface rendering algorithm in Salisbury & Tarr 1997.

    \author    Sonny Chan
    \date      January 2010, April 2014
*/
//===========================================================================

#include "ImplicitMesh.h"
#include "MarchingSource.h"

using namespace chai3d;


ImplicitMesh::ImplicitMesh()
    : m_projectedSphere(0.05)
{
    // because we are haptically rendering this object as an implicit surface
    // rather than a set of polygons, we will not need a collision detector
    setCollisionDetector(0);
    
    // create a sphere that tracks the position of the proxy
    m_projectedSphere.m_material->setWhite();
    addChild(&m_projectedSphere);
}

void ImplicitMesh::createFromFunction(double (*f)(double, double, double),
                            cVector3d a_lowerBound, cVector3d a_upperBound,
                            double a_granularity)
{
    // variables to hold raw triangles returned from marching cubes algorithm
    GLint tcount;
    GLfloat vertices[5*3*3];

    // sample the implicit surface by stepping through each dimension
    for (GLfloat x = a_lowerBound.x(); x <= a_upperBound.x(); x += a_granularity)
        for (GLfloat y = a_lowerBound.y(); y <= a_upperBound.y(); y += a_granularity)
            for (GLfloat z = a_lowerBound.z(); z <= a_upperBound.z(); z += a_granularity)
            {
                // call marching cubes to get the triangular facets for this cell
                vMarchCubeCustom(x, y, z, a_granularity, f, tcount, vertices);

                // add resulting triangles (if any) to our mesh
                for (int i = 0; i < tcount; ++i) {
                    int ix = i*9;
                    this->newTriangle(
                        cVector3d(vertices[ix+0], vertices[ix+1], vertices[ix+2]),
                        cVector3d(vertices[ix+3], vertices[ix+4], vertices[ix+5]),
                        cVector3d(vertices[ix+6], vertices[ix+7], vertices[ix+8])
                    );
                }
            }

    // compute face normals for our mesh so that lighting works properly
    this->computeAllNormals();
}

//! Contains code for graphically rendering this object in OpenGL.
void ImplicitMesh::render(cRenderOptions& a_options)
{
    // update the position and visibility of the proxy sphere
    m_projectedSphere.setShowEnabled(m_interactionInside);
    m_projectedSphere.setLocalPos(m_interactionPoint);
    
    // get the base class to render the mesh
    cMesh::render(a_options);
}


//===========================================================================
/*!
    This method should contain the core of the implicit surface rendering
    algorithm implementation.  The member variables m_interactionPoint
    and m_interactionInside should both be set by this method.

    \param  a_toolPos  Position of the tool.
    \param  a_toolVel  Velocity of the tool.
    \param  a_IDN  Identification number of the force algorithm.
*/
//===========================================================================
void ImplicitMesh::computeLocalInteraction(const cVector3d& a_toolPos,
                                           const cVector3d& a_toolVel,
                                           const unsigned int a_IDN)
{
    /////////////////////////////////////////////////////////////////////////
    // CS277 IMPLICIT SURFACE RENDERING ALGORITHM
    /////////////////////////////////////////////////////////////////////////
    
    // For Part II, the friction coefficients can be read from the object's
    // material as shown here.
    double mu_s = m_material->getStaticFriction();
    double mu_k = m_material->getDynamicFriction();
    
    
    // m_interactionProjectedPoint is the "proxy" point on the surface
    // that the rendering algorithm tracks.  It should be equal to the
    // tool position when the tool is not in contact with the object.
    m_interactionPoint = a_toolPos;

    // m_interactionInside should be set to true when the tool is in contact
    // with the object.
    m_interactionInside = false;
}
